import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';
import 'package:lori_sdk/data/database/interfaces/cartDao.dart';
import 'package:lori_sdk/data/lori_api.dart';
import 'package:lori_sdk/data/models/cart/cart.dart';
import 'package:lori_sdk/data/models/cart/cartItem.dart';
import 'package:lori_sdk/data/models/cart/freight.dart';
import 'package:lori_sdk/data/models/cart/installment.dart';
import 'package:lori_sdk/data/models/cart/paymentMethod.dart';
import 'package:lori_sdk/data/models/cart/seller.dart';
import 'package:lori_sdk/data/models/cart/sellerModality.dart';
import 'package:uuid/uuid.dart';

class SalesCommerceCartsApi {
  static const PARAMETER_CART_ID = '#cart_id#';

  static const GET_CART = 'sales/commerce/carts/$PARAMETER_CART_ID';

  static const UPDATE_CART_ITEM = 'sales/commerce/carts/$PARAMETER_CART_ID/items';

  static const UPDATE_GIFT_ITEM = 'sales/commerce/carts/$PARAMETER_CART_ID/items/gift';

  static const FINALIZE = 'sales/commerce/carts/$PARAMETER_CART_ID/finalize';

  static const GET_FREIGHTS = 'sales/commerce/carts/$PARAMETER_CART_ID/freights';

  static const GET_INSTALLMENTS = 'sales/commerce/carts/$PARAMETER_CART_ID/installments';

  static const GET_MODALITIES = 'sales/commerce/carts/$PARAMETER_CART_ID/modalities';

  static const GET_PAYMENT_METHODS = 'sales/commerce/carts/$PARAMETER_CART_ID/paymentmethods';

  static const GET_SELLERS = 'sales/commerce/carts/$PARAMETER_CART_ID/sellers';

  static const PARAMETER_SELLER_ID = '#seller_id#';
  static const GET_SELLERS_ITEMS = 'sales/commerce/carts/$PARAMETER_CART_ID/sellers/$PARAMETER_SELLER_ID/items';

  static const PARAMETER_ADDRESS_ID = '#address_id#';
  static const PUT_ADDRESS = 'sales/commerce/carts/$PARAMETER_CART_ID/address/$PARAMETER_ADDRESS_ID';

  static const PUT_BUYER = 'sales/commerce/carts/$PARAMETER_CART_ID/buyer';

  static const PARAMETER_CEP = '#cep#';
  static const PUT_CEP = 'sales/commerce/carts/$PARAMETER_CART_ID/cep/$PARAMETER_CEP';

  static const PARAMETER_CPF = '#cpf#';
  static const PUT_CPF_PICKUP = 'sales/commerce/carts/$PARAMETER_CART_ID/cpfpickup/$PARAMETER_CPF';

  static const PARAMETER_CVV = '#cvv#';
  static const PUT_CVV = 'sales/commerce/carts/$PARAMETER_CART_ID/cvv/$PARAMETER_CVV';

  static const PARAMETER_DEALER_CODE = '#dealer_code#';
  static const PUT_DEALER_CODE = 'sales/commerce/carts/$PARAMETER_CART_ID/dealercode/$PARAMETER_DEALER_CODE';

  static const PARAMETER_FINGERPRINT_TOKEN = '#fingerprint_token#';
  static const PUT_DEVICE_FINGERPRINT = 'sales/commerce/carts/$PARAMETER_CART_ID/devicefingerprint/$PARAMETER_FINGERPRINT_TOKEN';

  static const PARAMETER_FREIGHT_REFERENCE = '#freight_reference#';
  static const PUT_FREIGHT = 'sales/commerce/carts/$PARAMETER_CART_ID/freight/$PARAMETER_FREIGHT_REFERENCE';

  static const PARAMETER_NUMBER_INSTALLMENTS = '#number_installments#';
  static const PUT_INSTALLMENTS = 'sales/commerce/carts/$PARAMETER_CART_ID/installments/$PARAMETER_NUMBER_INSTALLMENTS';

  static const PUT_LOCATION = 'sales/commerce/carts/$PARAMETER_CART_ID/location';

  static const PARAMETER_MODALITY_ID = '#modality_id#';
  static const PUT_MODALITY = 'sales/commerce/carts/$PARAMETER_CART_ID/modality/$PARAMETER_MODALITY_ID';

  static const PARAMETER_PAYMENT_METHOD_ID = '#payment_method_id#';
  static const PUT_PAYMENT_METHOD = 'sales/commerce/carts/$PARAMETER_CART_ID/paymentmethod/$PARAMETER_PAYMENT_METHOD_ID';

//  static const PARAMETER_SELLER_ID = '#seller_id#'; (Already declared)
  static const PUT_SELLER = 'sales/commerce/carts/$PARAMETER_CART_ID/seller/$PARAMETER_SELLER_ID';

  static const PARAMETER_VOUCHER_CODE = '#voucher_code#';
  static const PUT_VOUCHER = 'sales/commerce/carts/$PARAMETER_CART_ID/voucher/$PARAMETER_VOUCHER_CODE';

  static const DELETE_VOUCHER = 'sales/commerce/carts/$PARAMETER_CART_ID/voucher';

  Dio _requester;

  SalesCommerceCartsApi({@required Dio requester}) : this._requester = requester;

  Future<ApiCallback<void>> validateCart() async {
    // Get the current cart
    Cart cart = await CartDao().get();
    // Check if we already have one
    if (cart != null) {
      //If yes just we need do update the cart
      try{
        String channel = LoriApi.operatingSystem;
        await getCart(cartId: cart.id, channel: channel);
        return ApiCallback(isSuccessful: true);
      }on DioError catch (e) {
        return ApiCallback<void>().defaultErrorCallback(e);
      }
    } else {
      // If no, make sure to create one
      ApiCallback createCartResult = await getCart();
      if (createCartResult.isSuccessful) {
        return ApiCallback(isSuccessful: true);
      } else {
        // We can't continue if we don't succeed on creating a cart
        return ApiCallback(isSuccessful: false);
      }
    }
  }

  Future<ApiCallback<Cart>> getCart({String cartId, String channel, bool saveAsDefault = true}) async {
    try {
      if (cartId == null) {
        cartId = await _getDefaultCartId();
        if (channel == null) channel = LoriApi.operatingSystem;
      }

      // Set custom channel header if needed
      var headers = Map<String, dynamic>();
      if (channel != null) {
        headers['channel'] = channel;
        headers['cartid'] = cartId;
      }
      var options = Options(responseType: ResponseType.json, headers: headers);

      var response = await _requester.get(GET_CART.replaceAll(PARAMETER_CART_ID, cartId), options: options);

      if (saveAsDefault) {
        // Parse the cart
        var parsedCart = Cart.fromJson(response.data);
        // Save it locally, if set to be default
        if (saveAsDefault) {
          await CartDao().insertOrUpdate(parsedCart);
        }
      }

      return ApiCallback<Cart>(isSuccessful: true, resultData: Cart.fromJson(response.data));
    } on DioError catch (e) {
      return ApiCallback<Cart>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> upgradeItem({@required String skuId, @required int quantity, bool isGift = false, String cartId, String channel, bool saveAsDefault = true}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var queryParameters = {
        'skuId': skuId,
        'quantity': quantity,
      };

      // Set custom channel header if needed
      var headers = Map<String, dynamic>();
      if (channel != null) {
        headers['channel'] = channel;
      }
      var options = Options(responseType: ResponseType.json, headers: headers);

      var response = await _requester.post((isGift ? UPDATE_GIFT_ITEM : UPDATE_CART_ITEM).replaceAll(PARAMETER_CART_ID, cartId), queryParameters: queryParameters, options: options);

      if (saveAsDefault) {
        // Parse the new cart
        var parsedCart = Cart.fromJson(response.data);
        // Check if we have a current cart
        var currentCart = await CartDao().get();
        // If we have a current cart, update the data based on the current existing cart
        if (currentCart != null) parsedCart.isUpdatingFrom(currentCart);
        // Save it locally
        await CartDao().insertOrUpdate(parsedCart);
      }

      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> downgradeItem({@required String skuId, @required int quantity, bool isGift = false, String cartId, String channel, bool saveAsDefault = true}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var queryParameters = {
        'skuId': skuId,
        'quantity': quantity,
      };

      // Set custom channel header if needed
      var headers = Map<String, dynamic>();
      if (channel != null) {
        headers['channel'] = channel;
      }
      var options = Options(responseType: ResponseType.json, headers: headers);

      var response = await _requester.delete((isGift ? UPDATE_GIFT_ITEM : UPDATE_CART_ITEM).replaceAll(PARAMETER_CART_ID, cartId), queryParameters: queryParameters, options: options);

      if (saveAsDefault) {
        // Parse the new cart
        var parsedCart = Cart.fromJson(response.data);
        // Check if we have a current cart
        var currentCart = await CartDao().get();
        // If we have a current cart, update the data based on the current existing cart
        if (currentCart != null) parsedCart.isUpdatingFrom(currentCart);
        // Save it locally
        await CartDao().insertOrUpdate(parsedCart);
      }

      return ApiCallback(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> finalize({String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(FINALIZE.replaceAll(PARAMETER_CART_ID, cartId));

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    } catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<List<Freight>>> getFreights({String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.get(GET_FREIGHTS.replaceAll(PARAMETER_CART_ID, cartId));
      // Parse result
      var responseData = response.data as List;
      var freights = List<Freight>();
      for (var row in responseData) {
        freights.add(Freight.fromJson(row));
      }

      // Callback invocation
      return ApiCallback<List<Freight>>(isSuccessful: true, resultData: freights);
    } on DioError catch (e) {
      return ApiCallback<List<Freight>>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<List<Installment>>> getInstallments({String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.get(GET_INSTALLMENTS.replaceAll(PARAMETER_CART_ID, cartId));
      // Parse result
      var responseData = response.data as List;
      var installments = List<Installment>();
      for (var row in responseData) {
        installments.add(Installment.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<Installment>>(isSuccessful: true, resultData: installments);
    } on DioError catch (e) {
      return ApiCallback<List<Installment>>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<List<SellerModality>>> getModalities({String latitude, String longitude, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      Map<String, String> queryParameters = {};

      if (latitude != null && longitude != null) {
        queryParameters['lat'] = latitude;
        queryParameters['lng'] = longitude;
      } else {
        queryParameters['lat'] = '0.0';
        queryParameters['lng'] = '0.0';
      }

      var response = await _requester.get(GET_MODALITIES.replaceAll(PARAMETER_CART_ID, cartId), queryParameters: queryParameters);
      // Parse result
      var responseData = response.data as List;
      var modalities = List<SellerModality>();
      for (var row in responseData) {
        modalities.add(SellerModality.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<SellerModality>>(isSuccessful: true, resultData: modalities);
    } on DioError catch (e) {
      return ApiCallback<List<SellerModality>>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<List<PaymentMethod>>> getPaymentMethods({String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.get(GET_PAYMENT_METHODS.replaceAll(PARAMETER_CART_ID, cartId));
      // Fetch payment methods
      var responseData = response.data as List;
      var payments = List<PaymentMethod>();
      for (var row in responseData) payments.add(PaymentMethod.fromJson(row));
      // Callback invocation
      return ApiCallback<List<PaymentMethod>>(isSuccessful: true, resultData: payments);
    } on DioError catch (e) {
      return ApiCallback<List<PaymentMethod>>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<List<Seller>>> getSellers({String latitude, String longitude, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      Map<String, String> queryParameters = {};

      if (latitude != null && longitude != null) {
        queryParameters['lat'] = latitude;
        queryParameters['lng'] = longitude;
      } else {
        queryParameters['lat'] = '0.0';
        queryParameters['lng'] = '0.0';
      }

      var response = await _requester.get(GET_SELLERS.replaceAll(PARAMETER_CART_ID, cartId), queryParameters: queryParameters);
      // Parse result
      var responseData = response.data as List;
      var sellers = List<Seller>();
      for (var row in responseData) {
        sellers.add(Seller.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<Seller>>(isSuccessful: true, resultData: sellers);
    } on DioError catch (e) {
      return ApiCallback<List<Seller>>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<List<CartItem>>> getSellerItems({String cartId, String sellerId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.get(GET_SELLERS_ITEMS.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_SELLER_ID, sellerId));
      // Parse result
      var responseData = response.data as List;
      var cartItems = List<CartItem>();
      for (var row in responseData) {
        cartItems.add(CartItem.fromJson(row));
      }
      // Callback invocation
      return ApiCallback<List<CartItem>>(isSuccessful: true, resultData: cartItems);
    } on DioError catch (e) {
      return ApiCallback<List<CartItem>>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putAddress({@required String addressId, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_ADDRESS.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_ADDRESS_ID, addressId));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putBuyer({String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_BUYER.replaceAll(PARAMETER_CART_ID, cartId));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putCep({@required String cep, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_CEP.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_CEP, cep));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putCPFPickup({@required String cpf, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_CPF_PICKUP.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_CPF, cpf));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putCvv({@required String cvv, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_CVV.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_CVV, cvv));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putDealerCode({@required String dealerCode, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_DEALER_CODE.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_DEALER_CODE, dealerCode));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putDeviceFingerprint({@required String fingerprintToken, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_DEVICE_FINGERPRINT.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_FINGERPRINT_TOKEN, fingerprintToken));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putFreight({@required String freightReference, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_FREIGHT.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_FREIGHT_REFERENCE, freightReference));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putInstallments({@required String installments, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_INSTALLMENTS.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_NUMBER_INSTALLMENTS, installments));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putLocation({@required String latitude, @required String longitude, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var queryParameters = {
        'lat': latitude,
        'lng': longitude,
      };

      var response = await _requester.put(PUT_LOCATION.replaceAll(PARAMETER_CART_ID, cartId), queryParameters: queryParameters);
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putModality({@required String modalityId, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_MODALITY.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_MODALITY_ID, modalityId));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putPaymentMethod({@required String paymentMethodId, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_PAYMENT_METHOD.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_PAYMENT_METHOD_ID, paymentMethodId));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putSeller({@required String sellerId, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_SELLER.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_SELLER_ID, sellerId));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> putVoucher({@required String voucherCode, String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.put(PUT_VOUCHER.replaceAll(PARAMETER_CART_ID, cartId).replaceAll(PARAMETER_VOUCHER_CODE, voucherCode));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<ApiCallback<void>> deleteVoucher({String cartId}) async {
    try {
      if (cartId == null) cartId = await _getDefaultCartId();

      var response = await _requester.delete(DELETE_VOUCHER.replaceAll(PARAMETER_CART_ID, cartId));
      // Parse the new cart
      var parsedCart = Cart.fromJson(response.data);
      // Save it locally
      await CartDao().insertOrUpdate(parsedCart);

      return ApiCallback<void>(isSuccessful: true);
    } on DioError catch (e) {
      return ApiCallback<void>().defaultErrorCallback(e);
    }
  }

  Future<String> _getDefaultCartId() async {
    var currentCart = await CartDao().get();

    if (currentCart != null) {
      return currentCart.id;
    } else {
      return Uuid().v4();
    }
  }
}
