import 'package:collection/collection.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:lori_sdk/data/models/product/product.dart';

part 'skuAttribute.g.dart';

//Sku Attribute object returned from the API
@JsonSerializable()
class SkuSpecificAttribute {
  SkuSpecificAttribute();

  String id;
  String skuId;
  SkuAttributeType type;
  String title;
  String value;
  String label;
  dynamic order;

  factory SkuSpecificAttribute.fromJson(Map<String, dynamic> json) => _$SkuSpecificAttributeFromJson(json);

  Map<String, dynamic> toJson() => _$SkuSpecificAttributeToJson(this);
}

//Sku Attribute available type
enum SkuAttributeType {
  @JsonValue(0)
  text,

  @JsonValue(1)
  color,

  @JsonValue(2)
  picker
}

//Types created to simplify the UI construction
class SkuAttribute {
  String title;
  SkuAttributeType type;
  dynamic order;
  List<SkuAttributeValue> values;

  void sortValues() {
    if (title.toLowerCase() == 'tamanho' || title.toLowerCase() == 'size') {
      this.values.sort((lv, rv) {
        //Compare the weight of both values
        var compare = lv.sizeSortWeight.compareTo(rv.sizeSortWeight);
        if (compare != 0) {
          //If the weights are different return the comparison result
          return compare;
        }
        else {
          //If compare is 0 this means the weights are equal,
          //so, use the default sort
          return lv.compareTo(rv);
        }
      });
    }
    else {
      this.values.sort((lv, rv) => lv.compareTo(rv));
    }
  }
}

class SkuAttributeValue implements Comparable<SkuAttributeValue> {
  String label;
  String value;
  bool isAvailable = true;

  SkuAttributeValue({this.label, this.value});

  @override
  String toString() {
    if (this.label != null && this.label.isNotEmpty) {
      return this.label;
    }
    else {
      return this.value;
    }
  }

  @override
  bool operator ==(Object other) => identical(this, other) ||
      other is SkuAttributeValue &&
          runtimeType == other.runtimeType &&
          this.label == other.label &&
          this.value == other.value;

  @override
  int get hashCode => label.hashCode ^ value.hashCode;

  int compareTo(SkuAttributeValue other) {
    //Check if the attributes values are convertable to int
    var thisInt = int.tryParse(this.toString());
    var otherInt = int.tryParse(other.toString());

    if (thisInt == null || otherInt == null) {
      //If both values are not int we must compare the strings
      return this.toString().compareTo(other.toString());
    }
    else {
      //Compare the ints
      return thisInt.compareTo(otherInt);
    }
  }

  // Reference array to sort values that refers to a SIZE
  Map<String, int> _sizeWeightReference = {
    'p': -1, 's': -1, //pequeno/small
    'm': 1, //medio/medium
    'g': 2, 'l': 2, //grande/large
    'x': 4, 'e': 4 //extra/Xtra
  };

  // Order the attributes by a pre-determined weight
  int get sizeSortWeight {
    int weight = 0;
    for (var i=0; i<this.value.length; i++) {
      var currentLetter = this.value[i].toLowerCase();
      var letterWeight = _sizeWeightReference[currentLetter] ?? 0;
      weight += letterWeight;
    }

    return weight;
  }
}


extension SkuAttributesListHelper on Iterable<Sku> {
  ///Retrieve all the existing attributes *values* along all skus
  ///
  /// This method will return all availables attributes, organized by title
  ///
  /// There are NO assurance that all combinations are available
  List<SkuAttribute> availableAttributes() {
    var skuList = this;

    //Get the title of the attributes, ordered by the attribute order
    var firstSkuAttrs = skuList.first.attributes;
    firstSkuAttrs.sort((la, ra) => la.order.compareTo(ra.order));
    var attrTitles = firstSkuAttrs.map((a) => a.title).toList();

    bool isValid = true;

    skuList.forEach((sku) {
      var skuAttrs = sku.attributes;
      skuAttrs.sort((la, ra) => la.order.compareTo(ra.order));
      var skuAttrTitles = skuAttrs.map((a) => a.title).toList();

      if (!ListEquality().equals(attrTitles, skuAttrTitles)) {
        //If any sku has a different sku attributes than the others we can't use the sku attributes
        isValid = false;
      }
    });

    if (!isValid) return null;

    //Get ALL attributes and values available
    var allAttributes = skuList.expand((sku) => sku.attributes);

    //Organize the attributes by title
    return allAttributes.organizedAttributes();
  }

  ///Return all the available values for an specific SKU based on a filter
  ///
  /// This method should be use to update the availability of SKU attributes after the user has selected one
  List<SkuAttributeValue> availableValuesForSkuAttributeWithTitle(String title, {Map<String, String> filters = null}) {
    //Filter the sku list according to the specified filter
    var filteredSkuList = this;
    if (filters != null) {
      for (var key in filters.keys) {
        filteredSkuList = filteredSkuList.where((sku) {
          return sku.attributes.where((a) {
            return a.title == key && a.value == filters[key];
          }).length > 0;
        }).toList();
      }
    }

    //Get all attributes with the specified title
    var allAttributes = filteredSkuList.expand((sku) => sku.attributes).where((at) => at.title == title);

    //Return the values
    return allAttributes.attributeWithTitle(title).values;
  }
}

extension SkuAttributeValuesHelper on Iterable<SkuSpecificAttribute> {
  /// "Organize" an array of SkuSpecificAttributes, grouping the sku values by title
  List<SkuAttribute> organizedAttributes() {
    //Get all titles available
    var allTitles = this.map((attr) => attr.title).toSet().toList();

    //Create the array to hold the available attributes
    List<SkuAttribute> organizedAttributes = List<SkuAttribute>();

    //Loop through all the labels
    allTitles.forEach((title) {
      var attributeValues = this.attributeWithTitle(title);

      organizedAttributes.add(attributeValues);
    });

    return organizedAttributes;
  }

  /// "Organize" an array of SkuSpecificAttributed, but considering only one title
  SkuAttribute attributeWithTitle(String title) {
    //Get all attributes with this label
    var attributesWithTitle = this.where((a) => a.title == title);

    //Create the wrapper
    var attributeValues = SkuAttribute();
    attributeValues.title = title;
    //The type should be equal on all attributes with the same label, so pick the first
    attributeValues.type = attributesWithTitle.first.type;
    //The same is valid to order
    attributeValues.order = attributesWithTitle.first.order;
    //Create an set with all the values
    attributeValues.values = attributesWithTitle.map((a) => SkuAttributeValue(label: a.label, value: a.value)).toList().toSet().toList();
    //Sort the values
    attributeValues.values.sort((lv, rv) => lv.toString().compareTo(rv.toString()));

    return attributeValues;
  }
}